// Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.
//
// This file is licensed according to the BSD 3-clause license as follows:
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the "Linz Center of Mechatronics GmbH" and "LCM" nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL "Linz Center of Mechatronics GmbH" BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// This file is part of X2C. http://x2c.lcm.at/
// $LastChangedRevision: 2465 $
// $LastChangedDate:: 2022-02-15 15:12:58 +0100#$
//
// Function to create 'builder.sce' for X2C libraries
// Parameters:
// libName - Library name
// varargin - (optional) project root directory for external blocks

function [] = generateLibraryHelp(libName, varargin)

// import necessary java classes
jimport at.lcm.x2c.utils.LibraryUtils;

// get XML file names
if length(varargin) == 0 then
    // internal library
    xmlPath = jinvoke(LibraryUtils, "getXmlDirectory", libName);
    libRoot = jinvoke(jinvoke(LibraryUtils, "getLibraryRootDirectory"), "toString");
else
    // external library
    libRoot = jinvoke(jinvoke(LibraryUtils, "getLibraryRootDirectory", varargin(1)), "toString");
    xmlPath = jinvoke(LibraryUtils, "getXmlDirectory", libName, libRoot);
end
xmlDir = jinvoke(xmlPath, "toString");
cur_lib_root = fullfile(libRoot, libName);

// get/create help directories
HelpDir = fullfile(getLibraryScilabPath(cur_lib_root), "help");
LangDir = fullfile(HelpDir, "en_US");
if ~isdir(HelpDir) then
    mkdir(HelpDir);
end
if ~isdir(LangDir) then
    mkdir(LangDir);
end

// get XML files
xmlFiles = dir(xmlDir + "*.xml");
xmlFiles = xmlFiles(2)';
xmlList = [];

// go through XML list
for curXML = xmlFiles
    if ~strcmp(part(curXML, $-3:$), ".xml", "i") then
        // add file name with *.xml ending
        xmlList = [xmlList, curXML];
    end
end

// create builder_help.sce file
err = createBuilderHelp(HelpDir);
if err ~= 0 then
    mprintf("Error: Couldn''t create builder_help.sce in %s\n!", HelpDir)
end

// create cleaner_help.sce file
err = createCleanerHelp(HelpDir);
if err ~= 0 then
    mprintf("Error: Couldn''t create cleaner_help.sce in %s\n!", HelpDir)
end

// create build_help.sce file
err = createBuildHelp(LangDir, libName);
if err ~= 0 then
    mprintf("Error: Couldn''t create build_help.sce in %s\n!", LangDir)
end

// create CHAPTER file
err = createChapter(LangDir, libName);
if err ~= 0 then
    mprintf("Error: Couldn''t create CHAPTER in %s\n!", LangDir)
end


// create docbook xml help files
for curXml=xmlList
    err = createXMLHelp(LangDir, curXml, cur_lib_root);
    if err ~= 0 then
        mprintf("Error: Couldn''t create docbook xml file in %s from %s\n!", LangDir, curXml)
    end
end


endfunction


function [err] = createXMLHelp(Dir, xmlFileName, lib_root)
// Function to create docbook xml help file.
//
// $LastChangedRevision: 2465 $
// 
// Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.

// import necessary java classes
jimport at.lcm.x2c.core.structure.ControlBlock;
jimport at.lcm.x2c.core.structure.Block;

try
    // read block XML file
    x2cBlock = jinvoke(Block, "loadBlockXML", xmlFileName);
    
    // get block info
    BlockName = jinvoke(x2cBlock, "getName");
    BlockPrompt = adjustString(jinvoke(x2cBlock, "getPrompt"));
    BlockDescription = adjustString(jinvoke(x2cBlock, "getDescription"));
    LibraryName = jinvoke(x2cBlock, "getLibraryName");
    MaskObj = jinvoke(x2cBlock, "getMask");
    InportNames = jinvoke(MaskObj, "getMaskInportNames");
    OutportNames = jinvoke(MaskObj, "getMaskOutportNames");
    MaskParameterNames = jinvoke(MaskObj, "getMaskParameterNames");
    icon_path = fullfile(getLibraryScilabPath(lib_root), 'Icons');

    // get implementation/datatype names
    blockType = jinvoke(jinvoke(x2cBlock, "getClass"), "toString");
    if ~strcmp(blockType, "class at.lcm.x2c.core.structure.ControlBlock") then // normal Block
        ImplNames = jinvoke(x2cBlock, "getExistingImplementationNames");
    else        // IO-Block
        ImplNames = getIODataTypes();
    end
    
    // check for empty strings
    if isempty(BlockPrompt) then
        BlockPrompt = "";
    end
    
    // open/create file
    fileID = mopen(fullfile(Dir, "x2c_" + BlockName + ".xml"), "wt");
            
    // write header
    mfprintf(fileID, "<?xml version='"1.0'" encoding='"UTF-8'"?>\n");
    mfprintf(fileID, "<refentry version='"5.0-subset Scilab'" xml:lang='"en'"\n");
    mfprintf(fileID, "    xmlns='"http://docbook.org/ns/docbook'"\n");
    mfprintf(fileID, "    xmlns:xlink='"http://www.w3.org/1999/xlink'"\n");
    mfprintf(fileID, "    xmlns:svg='"http://www.w3.org/2000/svg'"\n");
    mfprintf(fileID, "    xmlns:ns3='"http://www.w3.org/1999/xhtml'"\n");
    mfprintf(fileID, "    xmlns:mml='"http://www.w3.org/1998/Math/MathML'"\n");
    mfprintf(fileID, "    xmlns:db='"http://docbook.org/ns/docbook'">\n");
    mfprintf(fileID, "\n");
    
    // write id, name and prompt
    mfprintf(fileID, "    <refnamediv xml:id='"x2c_%s'">\n", BlockName);
    mfprintf(fileID, "        <refname>x2c_%s</refname>\n", BlockName);
    mfprintf(fileID, "        <refpurpose>%s</refpurpose>\n", BlockPrompt);
    mfprintf(fileID, "    </refnamediv>\n");
    mfprintf(fileID, "\n");
    
    // write block image
    mfprintf(fileID, "    <refsection>\n");
    mfprintf(fileID, "        <para>\n");
    icon_file = fullfile(icon_path, BlockName + ".png");
    if isfile(icon_file) then
        mfprintf(fileID, "            <inlinemediaobject>\n");
        mfprintf(fileID, "                <imageobject>\n");
        // get relative link to icon image and replace operating system file separator by '/'
        // ATTENTION: Use 'getrelativefilename' with caution because of a bug => see Scilab bug #16822
        icon_file_rel = strsubst(getrelativefilename(Dir, icon_file), filesep(), "/");
        mfprintf(fileID, "                <imagedata fileref='"%s'" align='"center'" valign='"middle'"/>\n", icon_file_rel);
        mfprintf(fileID, "                </imageobject>\n");
        mfprintf(fileID, "            </inlinemediaobject>\n");

    else
        mfprintf(fileID, "            Block image not available.\n");
    end
    mfprintf(fileID, "        </para>\n");
    mfprintf(fileID, "    </refsection>\n");
    mfprintf(fileID, "\n");
    
    // write description
    mfprintf(fileID, "    <refsection>\n");
    mfprintf(fileID, "        <title>Description</title>\n");
    mfprintf(fileID, "        <para>%s</para>\n", BlockDescription);
    //TODO: add link to PDF
    //mfprintf(fileID, "\n");
    //mfprintf(fileID, "        <para>Add here link to PDF help</para>\n");

    mfprintf(fileID, "    </refsection>\n");
    mfprintf(fileID, "\n");
    
    // write inports
    mfprintf(fileID, "    <refsection>\n");
    mfprintf(fileID, "        <title>Inports</title>\n");
    if isempty(InportNames) then
        mfprintf(fileID, "        <para>None</para>\n");
    else
        mfprintf(fileID, "        <simplelist type='"vert'">\n");
        for curName=InportNames
            Obj = jinvoke(MaskObj, "getMaskInport", curName);
            FriendlyName = adjustString(jinvoke(Obj, "getDescription"));
            mfprintf(fileID, "            <member><emphasis role='"bold'">%s:</emphasis> %s</member>\n", curName, FriendlyName);
        end
        mfprintf(fileID, "        </simplelist>\n");
    end
    

    mfprintf(fileID, "    </refsection>\n");
    mfprintf(fileID, "\n");

    // write outports
    mfprintf(fileID, "    <refsection>\n");
    mfprintf(fileID, "        <title>Outports</title>\n");
    if isempty(OutportNames) then
        mfprintf(fileID, "        <para>None</para>\n");
    else
        mfprintf(fileID, "        <simplelist type='"vert'">\n");
        for curName=OutportNames
            Obj = jinvoke(MaskObj, "getMaskOutport", curName);
            FriendlyName = adjustString(jinvoke(Obj, "getDescription"));
            mfprintf(fileID, "            <member><emphasis role='"bold'">%s:</emphasis> %s</member>\n", curName, FriendlyName);
        end
        mfprintf(fileID, "        </simplelist>\n");
    end
    mfprintf(fileID, "    </refsection>\n");
    mfprintf(fileID, "\n");

    // write mask parameters
    mfprintf(fileID, "    <refsection>\n");
    mfprintf(fileID, "        <title>Mask Parameters</title>\n");
   if isempty(MaskParameterNames) then
        mfprintf(fileID, "        <para>None</para>\n");
    else
        mfprintf(fileID, "        <simplelist type='"vert'">\n");
        for curName=MaskParameterNames
            Obj = jinvoke(MaskObj, "getMaskParameter", curName);
            FriendlyName = adjustString(jinvoke(Obj, "getDescription"));
            mfprintf(fileID, "            <member><emphasis role='"bold'">%s:</emphasis> %s</member>\n", curName, FriendlyName);
        end
        mfprintf(fileID, "        </simplelist>\n");
    end
    mfprintf(fileID, "    </refsection>\n");
    mfprintf(fileID, "\n");

    // write implementations
    mfprintf(fileID, "    <refsection>\n");
    mfprintf(fileID, "        <title>Implementations</title>\n");
    mfprintf(fileID, "\n");
    mfprintf(fileID, "        <simplelist type='"vert'">\n");
    for curImpl=ImplNames
        if ~strcmp(blockType, "class at.lcm.x2c.core.structure.ControlBlock") then // normal Block
            // get friendly implementation names
            ImplObj = jinvoke(x2cBlock, "getExistingImplementation", curImpl);
            ImplFriendlyName = adjustString(jinvoke(ImplObj, "getFriendlyName"));
        else        // IO-Block
            // get friendly implementation names
            //ToDo: fetch friendly name for io blocks
            ImplFriendlyName = "";
        end
        mfprintf(fileID, "            <member><emphasis role='"bold'">%s:</emphasis> %s</member>\n", curImpl, ImplFriendlyName);
    end
    mfprintf(fileID, "        </simplelist>\n");
    mfprintf(fileID, "    </refsection>\n");
    mfprintf(fileID, "\n");
    
    mfprintf(fileID, "</refentry>\n");
    
    // close file
    mclose(fileID);
    
    err = 0;
catch
    // some error creating the file occured
    err = 1;
    //disp("Error: " + lasterror());
end
endfunction

function [returnStr] = adjustString(inputStr)
    // Function to replace incompatible strings for xml document
    returnStr = strsubst(inputStr, "&", "&amp;");
    returnStr = strsubst(returnStr, "'"", "&quot;");
    returnStr = strsubst(returnStr, "''", "&apos;");
    returnStr = strsubst(returnStr, "<", "&lt;");
    returnStr = strsubst(returnStr, ">", "&gt;");
endfunction

function [err] = createBuildHelp(Dir, libName)
// Function to create 'build_help.sce' for building scilab help
//
// $LastChangedRevision: 2465 $
// 
// Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.

try
    // open/create file
    fileID = mopen(fullfile(Dir, "build_help.sce"), "wt");
    
    // print header
    time = datevec(datenum());
    mfprintf(fileID, "// This file was generated by %s on %02d-%02d-%04d %02d:%02d\n", "generateLibraryHelp.sci", time(3), time(2), time(1), time(4), time(5));
    mfprintf(fileID, "// This file is released under the 3-clause BSD license. See COPYING-BSD.\n\n");

    mfprintf (fileID, "tbx_build_help('"X2C %s Library'",get_absolute_file_path('"build_help.sce'"));\n", libName);
    
    // close file
    mclose(fileID);
    
    err = 0;
catch
    // some error creating the file occured
    err = 1;
end
endfunction

function [err] = createBuilderHelp(Dir)
// Function to create 'builder_help.sce' for building scilab help
//
// $LastChangedRevision: 2465 $
// 
// Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.

try
    // open/create file
    fileID = mopen(fullfile(Dir, "builder_help.sce"), "wt");
    
    // print header
    time = datevec(datenum());
    mfprintf(fileID, "// This file was generated by %s on %02d-%02d-%04d %02d:%02d\n", "generateLibraryHelp.sci", time(3), time(2), time(1), time(4), time(5));
    mfprintf(fileID, "// This file is released under the 3-clause BSD license. See COPYING-BSD.\n\n");

    
    mfprintf(fileID, "here = get_absolute_file_path('"builder_help.sce'");\n");
    mfprintf(fileID, "d = dir(here);\n");
    mfprintf(fileID, "d = d.name(d.isdir);\n");
    mfprintf(fileID, "la_LA = d(grep(d, '"/[a-z]{2}_[A-Z]{2}/'", '"r'"));  // filtering out ./images/ ./mml/, ...\n");
    mfprintf(fileID, "tbx_builder_help_lang(la_LA, here);\n");
    mfprintf(fileID, "clear here d la_LA\n");
    
    // close file
    mclose(fileID);
    
    err = 0;
catch
    // some error creating the file occured
    err = 1;
end
endfunction

function [err] = createCleanerHelp(Dir)
// Function to create 'cleaner_help.sce' for building scilab help
//
// $LastChangedRevision: 2465 $
// 
// Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.

try
    // open/create file
    fileID = mopen(fullfile(Dir, "cleaner_help.sce"), "wt");
    
    // print header
    time = datevec(datenum());
    mfprintf(fileID, "// This file was generated by %s on %02d-%02d-%04d %02d:%02d\n", "generateLibraryHelp.sci", time(3), time(2), time(1), time(4), time(5));
    mfprintf(fileID, "// This file is released under the 3-clause BSD license. See COPYING-BSD.\n\n");

    mfprintf(fileID, "function cleaner_help()\n");
    mfprintf(fileID, "    path = get_absolute_file_path('"cleaner_help.sce'");\n");
    mfprintf(fileID, "    langdirs = dir(path);\n");
    mfprintf(fileID, "    langdirs = langdirs.name(langdirs.isdir);\n");
    mfprintf(fileID, "\n");
    mfprintf(fileID, "    for l = 1:size(langdirs, '"*'")\n");
    mfprintf(fileID, "        masterfile = fullpath(path + filesep() + langdirs(l) + '"/master_help.xml'");\n");
    mfprintf(fileID, "        mdelete(masterfile);\n");
    mfprintf(fileID, "\n");
    mfprintf(fileID, "        jarfile = fullpath(path + '"/../jar/scilab_'" + langdirs(l) + '"_help.jar'");\n");
    mfprintf(fileID, "        mdelete(jarfile);\n");
    mfprintf(fileID, "\n");
    mfprintf(fileID, "        tmphtmldir = fullpath(path + '"/'" + langdirs(l) + '"/scilab_'" + langdirs(l) + '"_help'");\n");
    mfprintf(fileID, "        rmdir(tmphtmldir, '"s'");\n");
    mfprintf(fileID, "    end\n");
    mfprintf(fileID, "endfunction\n");
    mfprintf(fileID, "\n");
    mfprintf(fileID, "cleaner_help();\n");
    mfprintf(fileID, "clear cleaner_help;\n");
    
    
    // close file
    mclose(fileID);
    
    err = 0;
catch
    // some error creating the file occured
    err = 1;
end
endfunction

function [err] = createChapter(Dir, libName)
// Function to create 'CHAPTER' for building scilab help
//
// $LastChangedRevision: 2465 $
// 
// Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.

try
    // open/create file
    fileID = mopen(fullfile(Dir, "CHAPTER"), "wt");
    
    mfprintf (fileID, "title = %s", libName);
    
    // close file
    mclose(fileID);

    
    err = 0;
catch
    // some error creating the file occured
    err = 1;
end
endfunction
