// Copyright (c) 2017, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.
//
// This file is licensed according to the BSD 3-clause license as follows:
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the "Linz Center of Mechatronics GmbH" and "LCM" nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL "Linz Center of Mechatronics GmbH" BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// This file is part of X2C. http://x2c.lcm.at/
// $LastChangedRevision: 2508 $
// $LastChangedDate:: 2022-03-04 17:33:53 +0100#$
//
// X2C setup
// Parameters:
// varargin (optional): If true, X2C will fully rebuild

function [] = initX2C(varargin)

// get current directory
curDir = pwd();

// get file separator
fs = filesep();

// get operating system
os = getos();

// default: no setup error
setupError = %f;

// name of startup file
filename = "scilab.ini";

// check for type of installation request
deInstall = %f;
fullInstall = %f
if (length(varargin)>0) then
    if (varargin(1)==%f) then
        deInstall = %t;
    end
    if (varargin(1)==%t) then
        fullInstall = %t;
    end
end

if (deInstall==%f) then
    //--------------------------------------------------------------------------
    // X2C Installation
    //--------------------------------------------------------------------------
    mprintf("\n\nSetting up X2C\n==============\n")
    
    // get version info of scilab
    [vers, opt] = getversion();
    arch = opt(2);                      // scilab architecture

    scilab_ver = getversion("scilab");
    scilab_major_ver = scilab_ver(1);
    scilab_minor_ver = scilab_ver(2);
    
    if (~((scilab_major_ver == 5) & (scilab_minor_ver == 5)) & ~((scilab_major_ver == 6) & (scilab_minor_ver == 1))) then
        mprintf("You are using Scilab version ''%s'', which is currently not supported by X2C.\n", vers);
        mprintf("Please be aware that X2C may not work properly!\n");
        mprintf("Supported Scilab versions: 5.5.x, 6.1.x\n");
    end
    
    // get X2C paths
    // NOTE: it is assumed that setup is called from <X2C_ROOT>\System\Scilab\Scripts
    X2CRootPath = strsubst(curDir, [fs + "System" + fs + "Scilab" + fs + "Scripts"], "");
    X2CJavaPath = fullfile(X2CRootPath, "System", "Java");
    ScriptPath = curDir;
    
    // add Core JAR file to access Java classes required for basic setup
    javaclasspath(fullfile(X2CJavaPath, "Core.jar"));
    
    jimport("java.io.File");
    jimport("at.lcm.x2c.utils.Utils");
    try
        jinvoke(Utils, "setEnvironmentRootVariable", jnewInstance(File, X2CRootPath));
    catch
        mprintf("WARNING: X2C_Root environment variable could not be set!\n")
        mprintf("Please modify environment variable manually\n");
    end
    
    X2CLibraryPaths = [];
    libs = ls(fullfile(X2CRootPath, "Library"))';
    for curLib = libs
        if isfile(fullfile(X2CRootPath, "Library", curLib, "x2c_library.xml")) then
            X2CLibraryPaths = [X2CLibraryPaths, curLib];
        end
    end
    
    //--------------------------------------------------------------------------
    // Initialize X2C for setup
    //--------------------------------------------------------------------------
    jimport('at.lcm.x2c.utils.JavaUtils');
    
    jRootPath = jnewInstance(File, X2CRootPath);
    
    // X2C java class paths
    strClassPath = jinvoke(JavaUtils, 'getClassPath', jRootPath, %f, %t);
    classPaths = strsplit(strClassPath, ';');
    
    // add appropriate X2C Scilab JAR file dependend of Scilab version
    scilab_base_path = fullfile(X2CRootPath, 'System', 'Scilab');
    if (scilab_major_ver < 6) then
        scilab_classpath = fullfile(scilab_base_path, 'Java_Scilab5', 'dist', 'X2cScilab.jar');
    else
        scilab_classpath = fullfile(scilab_base_path, 'Java', 'dist', 'X2cScilab.jar');
    end
    classPaths($+1) = scilab_classpath;
    
    // X2C java library paths
    strLibraryPath = jinvoke(JavaUtils, 'getLibraryPath', jRootPath);
    libraryPaths = strsplit(strLibraryPath, ';');
    
    // load java class paths
    for curPath = classPaths
        javaclasspath(curPath);
    end

    // load java library paths
    for curPath = libraryPaths
        javalibrarypath(curPath);
    end
    
    //------------------------------------------------------------------------------
    // Full rebuild
    //------------------------------------------------------------------------------
    if (fullInstall==%t) then
        try
            sys_x2c_root = jinvoke(Utils, 'getRootDirectory');
            // because comparing a Java object against null is not possible in Scilab, the Java File exists method is probed
            // => if the object is not null, exists succeed, otherwise it throws an exception => catch in Scilab
            sys_x2c_root_uri = jinvoke(sys_x2c_root, 'toURI');
            sys_x2c_root_valid = %t;
        catch
            sys_x2c_root_valid = %f;
        end
        
        cur_x2c_root_uri = jinvoke(jnewInstance(File, X2CRootPath), 'toURI');
        // compare directories by URI because it does have a defined syntax regarding trailing path separator
        if (~sys_x2c_root_valid | ~sys_x2c_root_uri.equals(cur_x2c_root_uri)) then
            mprintf("Environment variable not up to date.\n");
            mprintf("Please restart Scilab & execute setup, with full install option set, again.\n");
            return;
        end
        
        // full rebuild of X2C requested
        mprintf("Setting up workspace for full rebuild...");
        try
            // load Xcos libraries
            loadXcosLibs();
    
            // load X2C scripts
            exec(fullfile(ScriptPath, "loadFunctions.sce"), -1);
            fs = filesep(); // restore file separator
            
            mprintf("DONE\n");
            
        catch
            mprintf("FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end
        
        // fetch architecture info in X2C format
        arch = getArchitecture();
        
        // clean up Scilab directories
        mprintf("+ Deleting files from previous setup...\n");
        try
            for curLib = X2CLibraryPaths
                library_dir = fullfile(X2CRootPath, "Library", curLib);
                scilabDir = fullfile(library_dir, "Scilab");
                // Library Scilab version specific directory
                scilab_spec_dir = getLibraryScilabPath(library_dir);
                if isdir(scilabDir) then
                    // only remove files in existing directories
                    if strcmp(curLib, "Basic") then
                        // ignore Basic library
                        mdelete(fullfile(scilabDir, "builder_" + arch + ".*"));
                        mdelete(fullfile(scilabDir, "*.sci"));
                        mdelete(fullfile(scilabDir, "*.c"));
                        mdelete(fullfile(scilabDir, "Palette_Icons_generated.txt"));
                    end
                end
                
                if isdir(scilab_spec_dir) then
                    mdelete(fullfile(scilab_spec_dir, "*.xpal"));
                    if strcmp(curLib, "Basic") then
                        mdelete(fullfile(scilab_spec_dir, "*_" + arch + ".*"));
                    else
                        // delete generated files from Basic library
                        mdelete(fullfile(scilab_spec_dir, "libBasic_" + arch + ".*"));
                        mdelete(fullfile(scilab_spec_dir, "loader_" + arch + ".*"));
                        mdelete(fullfile(scilab_spec_dir, "Library_is_compiled_for_" + arch + ".txt"));
                    end
                end
            end
            mprintf("- DONE\n");
        catch
            mprintf("- FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end
        
        // rebuild libraries
        mprintf("+ Rebuild X2C libraries...\n");
        try
            for curLib = X2CLibraryPaths
                    if strcmp(curLib, "Basic") then
                        // all but Basic
                        createLibrary(curLib);
                    else
                        // Basic
                        scilabDir = [X2CRootPath + fs + "Library" + fs + curLib + fs + "Scilab"];
                        chdir(scilabDir)
                        exec([scilabDir + fs + "builder_" + arch + ".sce"], -1);	// building is sufficient
                        fs = filesep(); // restore file separator
                        chdir(curDir)
                    end
            end
            mprintf("- DONE\n");
        catch
            mprintf("- FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end
        
        // populate X2C library manager with existing libraries
        mprintf("+ Populating X2C library manager...\n");
        try
            x2c_LibManager_populate();
            mprintf("- DONE\n");
        catch
            mprintf("- FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end
        
        // load palettes
        mprintf("+ Loading palettes...\n");
        try
            exec([ScriptPath + fs + "loadPalettes.sce"], -1);
            fs = filesep(); // restore file separator
            mprintf("- DONE\n");
        catch
            mprintf("- FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end
        
        // Generated Block images which can be used as (persistent) palette icons
        mprintf("+ Creating palette icons...\n");
        try
            createAllLibraryIcons();
            mprintf("- DONE\n");
        catch
            mprintf("- FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end

        // create documentation for libraries
        mprintf("+ Creating library help files...\n");
        try
            createAllLibraryHelps();
            mprintf("- DONE\n");
        catch
            mprintf("- FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            //setupError = %t; -> a documentation error should not fail the setup process
        end
        
        // create documentation for wrapper scripts
        wrapperDir = [X2CRootPath + fs + "System" + fs + "Scilab" + fs + "x2c_wrapper"];
        if isdir(wrapperDir) then
            mprintf("+ Creating documentation for wrapper scripts...\n");
            try
                exec(wrapperDir + fs + "installWrapper.sce",-1)
                fs = filesep(); // restore file separator
                mprintf("- DONE\n");
            catch
                mprintf("- FAILED\n");
                mprintf("ERROR: %s\n\n", lasterror());
                setupError = %t;
            end
        end
        
        
    else
        // normal build -> clean up a little
        // remove palette files
        mprintf("Deleting files from previous setup...");
        try
            exec('getLibraryScilabPath.sci', -1);
            for curLib = X2CLibraryPaths
                scilabDir = getLibraryScilabPath(fullfile(X2CRootPath, "Library", curLib));
                if isdir(scilabDir) then
                    // only remove files in existing directories
                    mdelete(fullfile(scilabDir, "*.xpal"));
                end
            end
            mprintf("DONE\n");
        catch
            mprintf("FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
            setupError = %t;
        end
        
    end

    if ~setupError then
        //----------------------------------------------------------------------
        // Write startup file
        //----------------------------------------------------------------------
        mprintf("Opening/creating startup file...");
        try
            jimport("at.lcm.x2c.utils.Version");
          
            // create file
            fileID = mopen(SCIHOME + fs + filename, 'wt');   // file will be overwritten if it already exists
            
            mprintf("DONE\n");
            
            // print header
            time = datevec(datenum());
            mfprintf(fileID, "// This file is part of X2C. http://x2c.lcm.at/\n");
            mfprintf(fileID, "// \n");
            mfprintf(fileID, "// Copyright (c) 2014, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/\n");
            mfprintf(fileID, "// All rights reserved.\n");
            mfprintf(fileID, "// \n");
            mfprintf(fileID, "// This file was generated\n");
            mfprintf(fileID, "// - by %s on %02d-%02d-%04d %02d:%02d\n", "setup.sce", time(3), time(2), time(1), time(4), time(5));
            mfprintf(fileID, "// - by X2C version %s\n", jinvoke(Version, "getVersion"));
            mfprintf(fileID, "// \n");
            mfprintf(fileID, "  \n");
            
            mprintf("  Writing path entries...");
            try
                mfprintf(fileID, "try\n");
                mfprintf(fileID, "    mprintf('"Start X2C\\n'");\n\n");
                mfprintf(fileID, "    // X2C Java class paths\n");
                for curPath = classPaths
                    mfprintf(fileID, "    javaclasspath('"%s'");\n", curPath);
                end
                mfprintf(fileID, "    \n");
        
                mfprintf(fileID, "    // X2C Java library paths\n");
                if isempty(libraryPaths)
                    mfprintf(fileID, "    // none\n");
                else
                    for curPath = libraryPaths
                        mfprintf(fileID, "    javalibrarypath('"%s'");\n", curPath);
                    end
                end
                mfprintf(fileID, "\n");
                
                mprintf("DONE\n");
            catch
                mprintf("FAILED\n");
                mprintf("ERROR: %s\n\n", lasterror());
            end
            
            // Output X2C version information
            mprintf("  Writing version info output...");
            try
                mfprintf(fileID, "    // Output X2C version information\n");
                mfprintf(fileID, "    jimport('"at.lcm.x2c.utils.Version'");\n");
                mfprintf(fileID, "    mprintf('"- Version: %%s\\n'", jinvoke(Version, '"getVersion'"));\n");
                mprintf("DONE\n");
            catch
                mprintf("FAILED\n");
                mprintf("ERROR: %s\n\n", lasterror());
            end
            
            mprintf("  Writing function loading entries...");
        
            try
                mfprintf(fileID, "\n");
                
                // load Xcos libraries
                mfprintf(fileID, "    loadXcosLibs();\n");
                mfprintf(fileID, "\n");
                
                // load X2C script functions
                mfprintf(fileID, "    exec('"%s'", -1);\n", [ScriptPath + fs + "loadFunctions.sce"]);
                mfprintf(fileID, "\n");
                
                mfprintf(fileID, "    mprintf('"+ Populating X2C library manager\\n'");\n");
                mfprintf(fileID, "    x2c_LibManager_populate();\n");
                mfprintf(fileID, "    mprintf('"- Done \\n'");\n");
                mfprintf(fileID, "\n");
                
                // load compiled X2C libraries needed for simulation
                if (os == "Windows") then
                    mfprintf(fileID, "    loadLibs();\n");
                    mfprintf(fileID, "\n");
                else
                    // if loadLibs() is executed in Linux a "scicos_free is missing" error will occur. But only when executed within the startup script. After startup, loadLibs() can be manually executed without any problems!
                    // TODO: add support for other OS
                end
                
                // load X2C block functions
                mfprintf(fileID, "    exec('"%s'", -1);\n", [ScriptPath + fs + "loadPalettes.sce"]);
                mfprintf(fileID, "\n");
        
                mprintf("DONE\n");
            catch
                mprintf("FAILED\n");
                mprintf("ERROR: %s\n\n", lasterror());
            end
            
            wrapperHelpDir = X2CRootPath + fs + "System" + fs + "Scilab" + fs + "x2c_wrapper" + fs + "jar";
            if isdir(wrapperHelpDir) then
                mprintf("  Writing wrapper help entries...");
                try
                            
                    mfprintf(fileID, "    add_help_chapter('"dummy'", '"%s'");\n", wrapperHelpDir);
                    mfprintf(fileID, "\n");
                                
                    mprintf("DONE\n");
                catch
                    mprintf("FAILED\n");
                    mprintf("ERROR: %s\n\n", lasterror());
                end
            end
            
//            // add X2C menu entry
//            mfprintf(fileID, "    x2c_MenuItems = ['"startCommunicatorRMI'", '"startBlockGenerator'"];\n");
//            mfprintf(fileID, "    addmenu('"X2C'", ['"Communicator'", '"BlockGenerator'"], list(0, '"x2c_MenuItems'"));\n");
//            mfprintf(fileID, "\n");
            
            // add Xcos Tools menu entries
            mfprintf(fileID, "    xcosAddToolsMenu('"start X2C Communicator'", '"startCommunicatorRMI();x2c_communicatorConnect();'");\n");
            mfprintf(fileID, "    xcosAddToolsMenu('"transform X2C Model'", '"x2c_ModelTransformation(''''set'''', 0, 0);'");\n");
            mfprintf(fileID, "    xcosAddToolsMenu('"create X2C Project Documentation'", '"x2c_ProjectDocumentation(''''set'''', 0, 0);'");\n");
            mfprintf(fileID, "    xcosAddToolsMenu('"start X2C BlockGenerator'", '"startBlockGenerator();'");\n");
            mfprintf(fileID, "\n");
            
            // Clean up of workspace
            mfprintf(fileID, "    // Clean up\n");
            mfprintf(fileID, "    clear('"Version'");\n");
            mfprintf(fileID, "\n");
            
            mfprintf(fileID, "catch\n");
            mfprintf(fileID, "    disp(['"ERROR: '" + lasterror()])\n");
            mfprintf(fileID, "end\n\n");
            
            // close file
            mprintf("Closing startup file...");
            try
                mclose(fileID);
                mprintf("DONE\n");
            catch
                mprintf("FAILED\n");
                mprintf("ERROR: %s\n\n", lasterror());
            end
            
        catch
            mclose(fileID); // try to close file
            mprintf("FAILED\n");
            mprintf("ERROR: %s\n\n", lasterror());
        end
        
        // create activation script
        exec("createActivationScript.sci", -1);
        createActivationScript(X2CRootPath);
        mprintf("\n\nFirst part of X2C setup finished, please restart Scilab to complete the setup!\n\n");
    else
        // error during setup occured -> do not write startup file
        mprintf("Error during setup -> no startup file created.\n")
    end
    
    // auto exit if full rebuild option enabled
    if (length(varargin)>0 & varargin(1)==%t) then
        mprintf("Scilab will be closed in 5 seconds. Good bye!");
        sleep(5000);
        exit();
    end
    
else
    //--------------------------------------------------------------------------
    // X2C Deinstallation
    //--------------------------------------------------------------------------
    mprintf("\n\nDeinstallation of X2C\n=====================\n")
    mprintf("Deleting startup file...");
    // delete init script to disable X2C startup
    mdelete(fullfile(SCIHOME, filename));
    mprintf("DONE\n");
    
    mprintf("Deinstallation finished. Please restart Scilab.\n")

end

endfunction
